class ResfoParsingError(ValueError):
    """
    Indicates an error occurred during reading of an res file.
    """

    pass


class ResfoWriteError(ValueError):
    """
    Indicates an error occurred during writing of an res file.
    """

    pass
